use axum::http::StatusCode;
use axum::response::IntoResponse;
use axum::Json;
use sqlx::Statement;

use crate::responses::registration::RegistrationResponse;
use crate::types::error_code::ErrorCode;

use super::authentication_error::AuthenticationError;
use super::registration_error::RegistrationError;
use super::ErrorResponse;

macro_rules! map_err {
    ($err:ident, $($type:path => $target:path),+) => {
        $(
            if $err.is::<$type>() {
                return $target($err.downcast().unwrap());
            }
        )*
    }
}

#[derive(Debug, thiserror::Error)]
pub enum ApiError {
    #[error("Registration Error")]
    RegistrationError(#[from] RegistrationError),

    #[error("Authentication Error")]
    AuthenticationError(#[from] AuthenticationError),

    #[error("Database Error")]
    DBError(#[from] sqlx::Error),

    #[error("Generic Error")]
    Generic(anyhow::Error),
}

impl From<anyhow::Error> for ApiError {
    fn from(err: anyhow::Error) -> Self {
        map_err!(err,
            sqlx::Error       => ApiError::DBError,
            RegistrationError => ApiError::RegistrationError
        );

        ApiError::Generic(err)
    }
}

impl IntoResponse for ApiError {
    fn into_response(self) -> axum::response::Response {
        match self {
            ApiError::RegistrationError(e) => e.into_response(),
            ApiError::AuthenticationError(e) => e.into_response(),
            ApiError::DBError(err) => {
                tracing::error!("{}", err.to_string());
                (
                    StatusCode::INTERNAL_SERVER_ERROR,
                    Json(ErrorResponse::new(
                        ErrorCode::Unknown,
                        "Database error! If you are the application owner please take a look at your application logs.",
                        None,
                    )),
                )
                    .into_response()
            },
            ApiError::Generic(err) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                Json(ErrorResponse::new(
                    ErrorCode::Unknown,
                    "Fatal error occured! If you are the application owner please take a look at your application logs.",
                    None,
                )),
            )
                .into_response(),
            _ => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
        }
    }
}
